#include <unistd.h>
#include <slp.h>
#include "iscsi-platform.h"
#include "iscsi-config.h"
#include "iscsi-slp-discovery.h"

#define ISCSI_SERVICE "service:iscsi:target"
#define MAX_IFACES 10
#define IP_ADDR_LEN 16
#define RECORD_LEN 2048
#define PREDICATE_LEN 2048
#define CRED_PREDICATE_LEN 512
#define NAME_PREDICATE_LEN 512
#define INITIATOR_NAME "/etc/initiatorname.iscsi"

static int record_begin;

typedef struct _InterfaceInfo {
    int iface_count;
    struct sockaddr_in iface_addr[MAX_IFACES];
} IfaceInfo;

typedef struct slp_portal {
    struct slp_portal *next;
    char *ip;
    char *port;
    char *tag;
} slp_portal_t;

typedef struct slp_target {
    struct slp_target *next;
    char *target_name;
    slp_portal_t *portals;
    slp_portal_t *tail;
} slp_target_t;

typedef struct slp_discovery_process {
    struct slp_discovery_process *next;
    slp_target_t *targets;
    slp_target_t *tail;
    iscsi_discovery_process_t *discovery;
} slp_discovery_process_t;

typedef struct slp_discovery_process_list {
    slp_discovery_process_t *head;
    slp_discovery_process_t *tail;
} slp_discovery_process_list_t;

typedef struct slp_discovery_pg_cookie {
    iscsi_discovery_process_t *discovery;
    char *target_name;
    char *ip;
    char *port;
} slp_discovery_pg_cookie_t;

slp_discovery_process_list_t *list = NULL;

void
add_slp_portal(slp_target_t * target, char *ip, char *port, char *tag)
{
    slp_portal_t *portal = (slp_portal_t *) malloc(sizeof (slp_portal_t));

    portal->ip = ip;
    portal->port = port;
    portal->tag = tag;
    portal->next = NULL;
    target->tail->next = portal;
    target->tail = portal;

    debugmsg(7, "slp portal %p added\n", portal);
}

void
add_slp_target(slp_discovery_process_t * process,
	       char *name, char *ip, char *port, char *tag)
{
    slp_target_t *target = NULL;
    slp_portal_t *portal = NULL;

    target = (slp_target_t *) malloc(sizeof (slp_target_t));
    portal = (slp_portal_t *) malloc(sizeof (slp_portal_t));

    target->target_name = name;
    target->portals = target->tail = portal;
    target->next = NULL;

    portal->ip = ip;
    portal->port = port;
    portal->tag = tag;
    portal->next = NULL;

    process->tail->next = target;
    process->tail = target;

    debugmsg(7, "slp target %p added\n", target);
}

void
add_slp_discovery_process(iscsi_discovery_process_t * discovery,
			  char *name, char *ip, char *port, char *tag)
{
    slp_discovery_process_t *process;
    slp_target_t *target;
    slp_portal_t *portal;

    process =
	(slp_discovery_process_t *) malloc(sizeof (slp_discovery_process_t));
    target = (slp_target_t *) malloc(sizeof (slp_target_t));
    portal = (slp_portal_t *) malloc(sizeof (slp_portal_t));

    process->discovery = discovery;
    process->targets = process->tail = target;
    process->next = NULL;

    target->target_name = name;
    target->portals = target->tail = portal;
    target->next = NULL;

    portal->ip = ip;
    portal->port = port;
    portal->tag = tag;
    portal->next = NULL;

    if (list) {
	list->tail->next = process;
	list->tail = process;
    } else {
	list = (slp_discovery_process_list_t *)
	    malloc(sizeof (slp_discovery_process_list_t));
	list->head = list->tail = process;
    }

    debugmsg(7, "slp discovery data buffered for discovery process: %p\n",
	     discovery);
}

void
save_slp_discovery_data(char *name, char *ip,
			char *port, char *tag,
			iscsi_discovery_process_t * discovery)
{
    slp_discovery_process_t *process = NULL;
    slp_target_t *target = NULL;
    int slp_discovery_process_exit = 0;
    int slp_target_exit = 0;

    if (list) {
	for (process = list->head; process; process = process->next) {
	    if (process->discovery == discovery) {
		slp_discovery_process_exit = 1;
		for (target = process->targets; target; target = target->next) {
		    if (strcmp(target->target_name, name) == 0) {
			slp_target_exit = 1;
			break;
		    }
		}
		break;
	    }
	}

	if (slp_discovery_process_exit) {
	    if (slp_target_exit)
		add_slp_portal(target, ip, port, tag);
	    else
		add_slp_target(process, name, ip, port, tag);
	} else
	    add_slp_discovery_process(discovery, name, ip, port, tag);
    } else {
	add_slp_discovery_process(discovery, name, ip, port, tag);
    }
}

void
send_target_record(slp_target_t * target, iscsi_discovery_process_t * discovery)
{
    slp_portal_t *portal;
    char record[RECORD_LEN];
    char *c = record;
    int len = 0;
    int result = 0;

    memset(record, 0, RECORD_LEN);
    if (!record_begin) {
	sprintf(c, "DTN=%s\n", target->target_name);
	record_begin = 1;
    } else {
	sprintf(c, "TN=%s\n", target->target_name);
    }
    len += strlen(c);
    c += strlen(c);

    for (portal = target->portals; portal; portal = portal->next) {
	sprintf(c, "TT=%s\n", portal->tag);
	len += strlen(c);
	c += strlen(c);

	sprintf(c, "TP=%s\n", portal->port);
	len += strlen(c);
	c += strlen(c);

	sprintf(c, "TA=%s\n", portal->ip);
	len += strlen(c);
	c += strlen(c);
    }
    sprintf(c, ";\n");
    len += strlen(c);

    debugmsg(7, "target record %s\n", record);

    result = write(discovery->pipe_fd, record, len);
    if (result < 0)
	logmsg(AS_ERROR,
	       "Error in sending discovery data for target %p by "
	       "discovery process %p\n", target, discovery);

}

void
free_slp_portals(slp_portal_t * portals)
{
    slp_portal_t *portal = NULL;

    while (portals) {
	portal = portals;

	if (portal->ip)
	    free(portal->ip);
	if (portal->port)
	    free(portal->port);
	if (portal->tag)
	    free(portal->tag);

	portals = portals->next;
	free(portal);
    }
}

void
free_slp_targets(slp_target_t * targets)
{
    slp_target_t *target = NULL;

    while (targets) {
	target = targets;
	if (target->target_name)
	    free(target->target_name);
	if (target->tail)
	    target->tail = NULL;
	if (target->portals)
	    free_slp_portals(target->portals);

	targets = targets->next;
	free(target);
    }
}

void
free_slp_discovery_process(slp_discovery_process_t * process)
{
    slp_discovery_process_t *p = NULL;
    slp_discovery_process_t *n = NULL;

    if (process) {
	if (process->targets) {
	    process->tail = NULL;
	    free_slp_targets(process->targets);
	}

	for (n = list->head, p = list->head; n; n = n->next) {
	    if (n == process) {
		if (n == p)	/* head */
		    list->head = process->next;
		else
		    p->next = n->next;

		if (n == list->tail) {
		    if (n == p)	/*  this is the only process in list */
			list->tail = list->head;
		    else
			list->tail = p;
		}
		process->discovery = NULL;
		free(process);
		break;
	    }
	    p = n;
	}
    }
}

void
send_slp_discovery_data(iscsi_discovery_process_t * discovery)
{
    slp_discovery_process_t *process = NULL;
    slp_target_t *target = NULL;

    if (list) {
	for (process = list->head; process; process = process->next) {
	    if (process->discovery == discovery) {
		for (target = process->targets; target; target = target->next) {
		    send_target_record(target, discovery);
		}
		write(discovery->pipe_fd, "!\n", sizeof ("!\n") - 1);
		record_begin = 0;
		break;
	    }
	}
    }
}

void
free_slp_discovery_data(iscsi_discovery_process_t * discovery)
{
    slp_discovery_process_t *process = NULL;

    if (list) {
	for (process = list->head; process; process = process->next) {
	    if (process->discovery == discovery) {
		free_slp_discovery_process(process);
		break;
	    }
	}
	if ((list->head == NULL) && (list->tail == NULL)) {
	    free(list);
	    list = NULL;
	}
    }
}

int
IfaceGetInfo(IfaceInfo * ifaceinfo, iscsi_discovery_process_t * discovery)
{
    struct sockaddr *sa;
    struct sockaddr_in *sin;
    struct ifreq ifrlist[MAX_IFACES];
    struct ifreq ifrflags;
    struct ifconf ifc;
    int fd;
    int i;

    ifc.ifc_len = sizeof (struct ifreq) * MAX_IFACES;
    ifc.ifc_req = ifrlist;

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd == -1) {
	logmsg(AS_ERROR,
	       "Failed to create socket, for getting all IP interfaces "
	       "through discovery process %p\n", discovery);
	return -1;
    }
    if (ioctl(fd, SIOCGIFCONF, &ifc) == -1) {
	logmsg(AS_ERROR,
	       "ioctl failure, in getting all IP interfaces through "
	       "discovery process %p\n", discovery);
	return -1;
    }
    memset(ifaceinfo, 0, sizeof (IfaceInfo));
    for (i = 0; i < ifc.ifc_len / sizeof (struct ifreq); i++) {
	sa = (struct sockaddr *) &(ifrlist[i].ifr_addr);
	if (sa->sa_family == AF_INET) {
	    /* Get interface flags */
	    memcpy(&ifrflags, &(ifrlist[i]), sizeof (struct ifreq));
	    if (ioctl(fd, SIOCGIFFLAGS, &ifrflags) == 0) {
		/* skip the loopback interfaces */
		if ((ifrflags.ifr_flags & IFF_LOOPBACK) == 0) {
		    sin = (struct sockaddr_in *) sa;
		    memcpy(&(ifaceinfo->iface_addr[ifaceinfo->iface_count]),
			   sin, sizeof (struct sockaddr_in));
		    ifaceinfo->iface_count++;
		}
	    }
	}
    }
    return 0;
}

char *
get_all_ip(iscsi_discovery_process_t * discovery)
{
    int i;
    IfaceInfo ifaceinfo;
    char *addrstr = NULL;

    if (IfaceGetInfo(&ifaceinfo, discovery) == 0) {
	addrstr = (char *) malloc(ifaceinfo.iface_count * IP_ADDR_LEN);
	memset(addrstr, 0, ifaceinfo.iface_count * IP_ADDR_LEN);

	for (i = 0; i < ifaceinfo.iface_count; i++) {
	    strcat(addrstr, inet_ntoa(ifaceinfo.iface_addr[i].sin_addr));
	    if (i + 1 < ifaceinfo.iface_count) {
		strcat(addrstr, ",");
	    }
	}
    }
    return addrstr;
}

SLPBoolean
SrvUrlCallback(SLPHandle hslp,
	       const char *srvurl,
	       unsigned short lifetime, int errcode, void *cookie)
{
    char *target_name, *targetname;
    char *ip;
    char *port, *port_id;
    char *tmp_url, *start_url = NULL, *attr_url;
    iscsi_discovery_process_t *discovery = NULL;
    char *addr;
    char *tag = NULL;

    if (errcode == SLP_OK) {
	debugmsg(7, "SLP url :%s,%i\n", srvurl, lifetime);
    }

    if (srvurl)
	start_url = strdup(srvurl);

    tmp_url = start_url;

    if (cookie)
	discovery = (iscsi_discovery_process_t *) cookie;

    if (tmp_url) {
	attr_url = strdup(tmp_url);
	if (strncmp(tmp_url, "service:iscsi:target://", 23) == 0) {
	    tmp_url += 22;
	    *tmp_url = '\0';
	    tmp_url++;
	    ip = tmp_url;

	    port = strchr(tmp_url, ':');
	    *port = '\0';
	    port++;
	    tmp_url = port;

	    target_name = strchr(tmp_url, '/');
	    *target_name = '\0';
	    target_name++;

	    tmp_url = target_name;

	    debugmsg(7,
		     "iSCSi Target: ip %s port %s target "
		     "name %s, discovered by discovery process %p\n",
		     ip, port, target_name, discovery);

	    /* slp query do not return any information related to
	     * target tag. For time being setting it to 1.
	     */
	    addr = (char *) malloc(strlen(ip) + 1);
	    memset(addr, 0, strlen(ip) + 1);
	    memcpy(addr, ip, strlen(ip));

	    port_id = (char *) malloc(strlen(port) + 1);
	    memset(port_id, 0, strlen(port) + 1);
	    memcpy(port_id, port, strlen(port));

	    targetname = (char *) malloc(strlen(target_name) + 1);
	    memset(targetname, 0, strlen(target_name) + 1);
	    memcpy(targetname, target_name, strlen(target_name));

	    tag = (char *) malloc(strlen("0") + 1);
	    memset(tag, 0, strlen("0") + 1);
	    strcpy(tag, "1");
	    save_slp_discovery_data(targetname, addr, port_id, tag, discovery);
	    if (discovery->flag == SLP_UNICAST_ENABLED)
		slp_unicast_attr_query(discovery, attr_url, "portal-group",
				       targetname, addr, port_id);
	    if (discovery->flag == SLP_MULTICAST_ENABLED)
		slp_multicast_attr_query(discovery, attr_url, "portal-group",
					 targetname, addr, port_id);
	}
	free(start_url);
    } else {
	if (discovery) {
	    send_slp_discovery_data(discovery);
	    free_slp_discovery_data(discovery);
	}
    }
    return SLP_TRUE;
}

SLPBoolean
AttrQueryCallback(SLPHandle hslp,
		  const char *attrlist, int errcode, void *cookie)
{
    slp_discovery_process_t *process;
    slp_target_t *target;
    slp_portal_t *portal;
    iscsi_discovery_process_t *discovery = NULL;
    char *tmp_url, *start_url = NULL;
    char *tag, *tmp;
    char *ip = NULL, *port = NULL, *target_name = NULL;

    if (errcode == SLP_OK) {
	debugmsg(7, "SLP attribute list :%s\n", attrlist);
    }

    if (attrlist) {
	start_url = strdup(attrlist);
    }

    tmp_url = start_url;
    if (cookie) {
	discovery = (iscsi_discovery_process_t *) ((slp_discovery_pg_cookie_t *)
						   cookie)->discovery;
	ip = (char *) ((slp_discovery_pg_cookie_t *) cookie)->ip;
	port = (char *) ((slp_discovery_pg_cookie_t *) cookie)->port;
	target_name =
	    (char *) ((slp_discovery_pg_cookie_t *) cookie)->target_name;
    }
    if (tmp_url) {
	tmp = strstr(tmp_url, "portal-group=");
	tmp = strchr(tmp, '=');
	tmp++;
	tmp = strtok(tmp, ")");
	tag = strdup(tmp);
	if (list) {
	    for (process = list->head; process; process = process->next) {
		if (process->discovery == discovery) {
		    for (target = process->targets; target;
			 target = target->next) {
			if (strcmp(target->target_name, target_name) == 0) {
			    for (portal = target->portals; portal;
				 portal = portal->next) {
				if ((!strcmp(portal->ip, ip))
				    && (!strcmp(portal->port, port))) {

				    portal->tag = tag;
				}
			    }
			}
		    }
		}
	    }
	}
	free(start_url);
    } else {
	return SLP_TRUE;
    }
    return SLP_TRUE;
}

void
slp_multicast_srv_query(iscsi_discovery_process_t * discovery, char *pl,
			unsigned short flag)
{
    SLPError result;
    SLPHandle hslp;
    iscsi_slp_config_t *config = discovery->entry->config.slp;

    if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK) {

	if ((result = SLPAssociateIFList(hslp, config->interfaces)) != SLP_OK) {
	    logmsg(AS_ERROR,
		   "SLP query error on discovery %p with SLP errorcode: %i\n",
		   discovery, result);
	    SLPClose(hslp);
	    return;
	}
    }

    switch (flag) {

    case GENERIC_QUERY:
	result = SLPFindSrvs(hslp,
			     ISCSI_SERVICE,
			     NULL, pl, SrvUrlCallback, discovery);
	break;

    default:
	result = SLP_FALSE;
    }

    if (result != SLP_OK) {
	logmsg(AS_ERROR,
	       "SLP query error on discovery %p with SLP errorcode: %i\n",
	       discovery, result);
    }
    SLPClose(hslp);
}

void
slp_multicast_attr_query(iscsi_discovery_process_t * discovery, char *url,
			 char *pl, char *target_name, char *ip, char *port)
{
    SLPError result;
    SLPHandle hslp;
    iscsi_slp_config_t *config = discovery->entry->config.slp;
    slp_discovery_pg_cookie_t cookie;

    if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK) {

	if ((result = SLPAssociateIFList(hslp, config->interfaces)) != SLP_OK) {
	    logmsg(AS_ERROR,
		   "SLP query error on discovery %p with SLP errorcode: %i\n",
		   discovery, result);
	    SLPClose(hslp);
	    return;
	}
    }

    cookie.discovery = (iscsi_discovery_process_t *) discovery;
    cookie.target_name = (char *) malloc(strlen(target_name) + 1);
    memset(cookie.target_name, 0, strlen(target_name) + 1);
    strcpy(cookie.target_name, target_name);

    cookie.ip = (char *) malloc(strlen(ip) + 1);
    memset(cookie.ip, 0, strlen(ip) + 1);
    strcpy(cookie.ip, ip);

    cookie.port = (char *) malloc(strlen(port) + 1);
    memset(cookie.port, 0, strlen(port) + 1);
    strcpy(cookie.port, port);

    result = SLPFindAttrs(hslp, url, NULL, pl, AttrQueryCallback, &cookie);

    if (result != SLP_OK) {
	logmsg(AS_ERROR,
	       "SLP query error on discovery %p with SLP errorcode: %i\n",
	       discovery, result);
    }
    SLPClose(hslp);
}

void
slp_unicast_srv_query(iscsi_discovery_process_t * discovery, char *pl,
		      unsigned short flag)
{
    SLPError result;
    SLPHandle hslp;
    iscsi_slp_config_t *config = discovery->entry->config.slp;

    if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK) {
	if ((result = SLPAssociateIP(hslp, config->address)) != SLP_OK) {
	    logmsg(AS_ERROR,
		   "SLP query error on discovery %p with SLP errorcode: %i\n",
		   discovery, result);
	    SLPClose(hslp);
	    return;
	}
    }

    switch (flag) {

    case GENERIC_QUERY:
	result = SLPFindSrvs(hslp,
			     ISCSI_SERVICE,
			     NULL, pl, SrvUrlCallback, discovery);
	break;

    default:
	result = SLP_FALSE;
    }

    if (result != SLP_OK)
	logmsg(AS_ERROR,
	       "SLP query error on discovery %p with SLP errorcode: %i\n",
	       discovery, result);

    SLPClose(hslp);
}

void
slp_unicast_attr_query(iscsi_discovery_process_t * discovery, char *url,
		       char *pl, char *target_name, char *ip, char *port)
{
    SLPError result;
    SLPHandle hslp;
    iscsi_slp_config_t *config = discovery->entry->config.slp;
    slp_discovery_pg_cookie_t cookie;

    if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK) {
	if ((result = SLPAssociateIP(hslp, config->address)) != SLP_OK) {
	    logmsg(AS_ERROR,
		   "SLP query error on discovery %p with SLP errorcode: %i\n",
		   discovery, result);
	    SLPClose(hslp);
	    return;
	}
    }

    cookie.discovery = (iscsi_discovery_process_t *) discovery;
    cookie.target_name = (char *) malloc(strlen(target_name) + 1);
    memset(cookie.target_name, 0, strlen(target_name) + 1);
    strcpy(cookie.target_name, target_name);

    cookie.ip = (char *) malloc(strlen(ip) + 1);
    memset(cookie.ip, 0, strlen(ip) + 1);
    strcpy(cookie.ip, ip);

    cookie.port = (char *) malloc(strlen(port) + 1);
    memset(cookie.port, 0, strlen(port) + 1);
    strcpy(cookie.port, port);

    result = SLPFindAttrs(hslp, url, NULL, pl, AttrQueryCallback, &cookie);

    if (result != SLP_OK)
	logmsg(AS_ERROR,
	       "SLP query error on discovery %p with SLP errorcode: %i\n",
	       discovery, result);

    SLPClose(hslp);
}

char *
generate_pl(char *interfaces)
{
    char slp_predicate_list[PREDICATE_LEN];
    char *pl = slp_predicate_list;
    int SLPpl = 0;
    char *temp, *startp = strdup(interfaces);
    char *endp;
    char *c;

    temp = startp;

    if (strlen(temp)) {
	sprintf(pl, "(auth-addr=any)");

	endp = strrchr(temp, ',');	/* more than one IP address */
	if (endp) {
	    while (endp > temp) {
		c = strchr(temp, ',');
		*c = '\0';
		c += 1;
		sprintf(pl + strlen(pl), "(auth-addr=%s)", temp);
		temp = c;
	    }
	}
	sprintf(pl + strlen(pl), "(auth-addr=%s)", temp);
	free(startp);
	SLPpl = 1;
    } else
	SLPpl = 0;

    if (SLPpl) {
	pl = strdup(slp_predicate_list);
	return pl;
    }
    return NULL;
}

char *
multicast_predicate_list(iscsi_discovery_process_t * discovery)
{
    iscsi_slp_config_t *config = discovery->entry->config.slp;
    char *mpl = NULL;

    if (config->interfaces && strlen(config->interfaces)) {
	if (strcmp(config->interfaces, "none") == 0) {
	    debugmsg(1, "SLP Multicast is disabled for discovery process %p\n",
		     discovery);
	} else if (strcmp(config->interfaces, "all") == 0) {
	    free(config->interfaces);
	    config->interfaces = get_all_ip(discovery);
	    if (config->interfaces)
		mpl = generate_pl(config->interfaces);
	} else {		/* comma separated list of IP's */
	    mpl = generate_pl(config->interfaces);
	}
    }
    return mpl;
}

char *
unicast_predicate_list(iscsi_discovery_process_t * discovery)
{
    iscsi_slp_config_t *config = discovery->entry->config.slp;
    char *interfaces = NULL;
    char *upl = NULL;

    if (config->address && strlen(config->address)) {
	interfaces = get_all_ip(discovery);
	if (interfaces)
	    upl = generate_pl(interfaces);
    }

    return upl;
}

char *
cred_predicate_list(iscsi_discovery_process_t * discovery)
{
    iscsi_slp_config_t *config = discovery->entry->config.slp;
    char cred_predicate_list[CRED_PREDICATE_LEN];
    char *slpc = cred_predicate_list;

    sprintf(slpc, "(auth-cred=any)");

    if (strlen(config->auth_options.username)) {
	/* We ll have one more case when we add SRP support
	 */
	switch (config->auth_options.authmethod) {

	case CHAP_AUTHENTICATION:
	default:
	    sprintf(slpc + strlen(slpc), "(auth-cred=chap/%s)",
		    config->auth_options.username);
	}
    }

    slpc = strdup(cred_predicate_list);

    return slpc;
}

char *
name_predicate_list(void)
{
    char name_predicate[NAME_PREDICATE_LEN];
    char *slpn = name_predicate;
    char *initiator_name = NULL;

    sprintf(slpn, "(auth-name=any)");
    initiator_name = get_iscsi_initiatorname(INITIATOR_NAME);

    if (initiator_name && strlen(initiator_name)) {
	sprintf(slpn + strlen(slpn), "(auth-name=%s)", initiator_name);
    }

    slpn = strdup(name_predicate);

    return slpn;
}

char *
generate_predicate_list(iscsi_discovery_process_t * discovery,
			unsigned short *flag)
{
    char *mpl;
    char *upl;
    char *cpl;
    char *npl;
    char *pl;
    int plen = 0;

    mpl = multicast_predicate_list(discovery);
    upl = unicast_predicate_list(discovery);
    cpl = cred_predicate_list(discovery);
    npl = name_predicate_list();

    if (mpl) {
	debugmsg(7, "multicast predicate list for discovery process %p is %s\n",
		 discovery, mpl);
	plen += strlen(mpl);
	*flag = SLP_MULTICAST_ENABLED;
    }

    if (upl) {
	debugmsg(7, "unicast predicate list for discovery process %p is %s\n",
		 discovery, upl);
	plen += strlen(upl);
	*flag = SLP_UNICAST_ENABLED;
    }

    if (plen) {
	if (cpl) {
	    debugmsg(7,
		     "credential predicate list for discovery "
		     "process %p is %s\n", discovery, cpl);
	    plen += strlen(cpl);
	}
	if (npl) {
	    debugmsg(7, "name predicate list for discovery process %p is %s\n",
		     discovery, npl);
	    logmsg(AS_ERROR,
		   "name predicate list for discovery process %p is %s\n",
		   discovery, npl);
	    plen += strlen(npl);
	}
    }

    if (plen) {
	char *c;

	plen += 16;		/* This extra space is required for 
				 * "(&(|%s)(|%s)(|%s)(|%s))"
				 * whearas, 1st %s is for mpl,
				 *          2nd %s is for upl,
				 *          3rd %s is for cpl,
				 *          4th %s is for npl.
				 */
	pl = (char *) malloc(sizeof (char) * plen);
	memset(pl, 0, plen);
	c = pl;

	sprintf(c, "(&");
	c += strlen(c);

	if (mpl) {
	    sprintf(c, "(|%s)", mpl);
	    c += strlen(c);
	    free(mpl);
	}
	if (upl) {
	    sprintf(c, "(|%s)", upl);
	    c += strlen(c);
	    free(upl);
	}

	sprintf(c, "(|%s)", cpl);
	c += strlen(c);
	free(cpl);

	sprintf(c, "(|%s)", npl);
	c += strlen(c);
	free(npl);

	sprintf(c, ")");	/* close brace correspond to '(&' */

	debugmsg(3, "SLP predicate list for discover process %p is %s\n",
		 discovery, pl);

	return pl;
    }

    return NULL;
}
