/*
 * iSCSI connection daemon
 * Copyright (C) 2002 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-network-boot.c,v 1.7.2.1 2004/08/10 23:04:49 coughlan Exp $
 *
 */

#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/utsname.h>

#include "iscsi-common.h"
#include "iscsi-ioctl.h"
#include "iscsid.h"
#include "iscsi-login.h"

#define ISCSI_DEVICE "/dev/iscsictl"

static int control_fd = -1;

extern void establish_iscsi_network_boot_session(int bus_id, int target_id,
						 struct sapiNBP *nbp);
int find_pxe_sig(char *, int, struct sapiNBP *);

int
open_control_device(void)
{
    FILE *f = NULL;
    char devname[256];
    char buf[256];
    int devn;
    int ctlfd;

    /*
     * Establish the control path to the driver
     */
    f = fopen("/proc/devices", "r");
    if (!f) {
	printf("Cannot open control path to the driver");
	return -1;
    }

    devn = 0;
    while (!feof(f)) {
	if (!fgets(buf, sizeof (buf), f)) {
	    break;
	}
	if (sscanf(buf, "%d %s", &devn, devname) != 2) {
	    continue;
	}
	if (!strcmp(devname, "iscsictl")) {
	    break;
	}
	devn = 0;
    }

    fclose(f);
    if (!devn) {
	printf("cannot find iscsictl in /proc/devices - make sure the "
	       "module is loaded");
	return -1;
    }

    unlink(ISCSI_DEVICE);
    if (mknod(ISCSI_DEVICE, (S_IFCHR | 0600), (devn << 8))) {
	printf("cannot create " ISCSI_DEVICE);
	return -1;
    }

    ctlfd = open(ISCSI_DEVICE, O_RDWR);
    if (ctlfd < 0) {
	printf("cannot open " ISCSI_DEVICE);
	return -1;
    }

    return ctlfd;
}

char *
get_iscsi_initiatorname(char *pathname)
{
    FILE *f = NULL;
    int c;
    char *line, buffer[1024];
    char *name = NULL;

    if (!pathname) {
	printf("No pathname to load InitiatorName from");
	return NULL;
    }

    /* get the InitiatorName */
    if ((f = fopen(pathname, "r"))) {
	while ((line = fgets(buffer, sizeof (buffer), f))) {

	    while (line && isspace(c = *line))
		line++;

	    if (strncmp(line, "InitiatorName=", 14) == 0) {
		char *end = line + 14;

		/* the name is everything up to the first bit of whitespace */
		while (*end && (!isspace(c = *end)))
		    end++;

		if (isspace(c = *end))
		    *end = '\0';

		if (end > line + 14)
		    name = strdup(line + 14);
	    }
	}
	fclose(f);
	if (!name) {
	    printf("an InitiatorName is required, but was not found in %s",
		   pathname);
	    return NULL;
	} else {
	    printf("InitiatorName=%s\n", name);
	}
	return name;
    } else {
	printf("cannot open InitiatorName configuration file %s", pathname);
	return NULL;
    }
}

/*
 * Assumptions for network boot:
 * 1. Only one portal available which comes from nbp structure so no
 *    MULTIPATHING and no portal failover.
 * 2. No authentication is used.
 * 3. Digest values are not configurable on initiator side, user can set it on 
 *    target if required.
 * 4. Tag is always 1, as there is no value in nbp specifying portal tag.
 * 5. Initiator name for boot and normal sessions will be different(FIXME).
 */

/* Returns 0 on success, non zero on failure */
static int
ioctl_establish_session(int bus_id, int target_id, struct sapiNBP *nbp)
{
    int rc, ret = 0;
    int p = 0;
    int probe_luns = 1;
    iscsi_session_ioctl_t *ioctld = NULL;

    /* allocate an ioctl structure with enough space for all
     * of the portal info 
     */
    ioctld = calloc(1, sizeof (*ioctld) + sizeof (iscsi_portal_info_t));
    if (ioctld == NULL) {
	printf("failed to allocate %d bytes for ioctl data structure",
	       (int) (sizeof (*ioctld) + sizeof (iscsi_portal_info_t)));
	return 1;
    }

    ioctld->ioctl_size = sizeof (*ioctld);
    ioctld->ioctl_version = ISCSI_SESSION_IOCTL_VERSION;
    ioctld->config_number = 1;
    ioctld->update = 0;
    ioctld->portal_info_size = sizeof (iscsi_portal_info_t);
    ioctld->num_portals = 1;

    p = 0;
    ioctld->portals[p].login_timeout = 15;
    ioctld->portals[p].auth_timeout = 45;
    ioctld->portals[p].active_timeout = 5;
    ioctld->portals[p].idle_timeout = 60;
    ioctld->portals[p].ping_timeout = 5;
    ioctld->portals[p].replacement_timeout = 0;
    ioctld->portals[p].disk_command_timeout = 0;	/* no maximum enforced */
    ioctld->portals[p].abort_timeout = 10;
    ioctld->portals[p].reset_timeout = 30;

    ioctld->portals[p].InitialR2T = 0;
    ioctld->portals[p].ImmediateData = 0;
    ioctld->portals[p].MaxRecvDataSegmentLength = 128 * 1024;
    ioctld->portals[p].FirstBurstLength = 256 * 1024;
    ioctld->portals[p].MaxBurstLength = (16 * 1024 * 1024) - 1024;
    ioctld->portals[p].DefaultTime2Wait = 0;	/* we only use
						 * session reinstatement
						 * (ERL 0) 
						 */
    ioctld->portals[p].DefaultTime2Retain = 0;	/* we only use 
						 * session reinstatement 
						 * (ERL 0) 
						 */
    ioctld->portals[p].DataDigest = ISCSI_DIGEST_NONE_CRC32C;
    ioctld->portals[p].HeaderDigest = ISCSI_DIGEST_NONE_CRC32C;
    ioctld->portals[p].tcp_window_size = 256 * 1024;
    ioctld->portals[p].ip_length = 4;
    memcpy(ioctld->portals[p].ip_address, (char *) (&nbp->targetipaddr), 4);

    ioctld->portals[p].port = nbp->tcpport;
    ioctld->portals[p].tag = PORTAL_GROUP_TAG_UNKNOWN;

    ioctld->password_length = 0;
    ioctld->password_length_in = 0;
    printf("authentication not configured for target");

    memset(&ioctld->lun_bitmap, 0xFF, sizeof (ioctld->lun_bitmap));

    strcpy(ioctld->InitiatorName,
	   get_iscsi_initiatorname("/etc/initiatorname.iscsi"));

    strcpy(ioctld->TargetName, nbp->targetstring);

    ioctld->isid[0] = DRIVER_ISID_0;
    ioctld->isid[1] = DRIVER_ISID_1;
    ioctld->isid[2] = DRIVER_ISID_2;
    ioctld->isid[3] = 0;
    ioctld->isid[4] = 0;
    ioctld->isid[5] = 1;

    ioctld->iscsi_bus = bus_id;
    ioctld->target_id = target_id;

    ioctld->probe_order = 0;
    ioctld->host_number = -1;
    ioctld->channel = -1;
    ioctld->portal_failover = 0;

    if (probe_luns)
	/* tell the kernel to probe luns */
	ioctld->probe_luns = 1;

    /*
     * Tell the kernel to establish a session.
     */
    do {
	rc = ioctl(control_fd, ISCSI_ESTABLISH_SESSION, (char *) ioctld);
	if (rc <= 0) {
	    printf("iSCSI session ioctl failed for %s, errno = %d",
		   nbp->targetstring, errno);
	    printf("\nRetrying session establishment with the boot disk ...\n");
	    sleep(2);
	} else {
	    printf("iSCSI session ioctl for %s returned from kernel with rc %d",
		   nbp->targetstring, rc);
	}

    } while (rc <= 0);

    /* we only get the host and channel numbers if a new session claimed them,
     * so this tells us whether we updated an existing session or started a new
     * one.  If we started a new one, log it, and maybe run a script.
     */
    if ((ioctld->host_number >= 0) && (ioctld->channel >= 0)) {
	printf("iSCSI: after ISCSI_ESTABLISH_SESSION host %d channel %d id "
	       "%d is target %s, alias %s",
	       ioctld->host_number, ioctld->channel, ioctld->target_id,
	       nbp->targetstring, ioctld->TargetAlias);
	/* FIXME: call a target-available script/program? */
    }

    /* don't want to leave passwords in memory */
    memset(&ioctld, 0, sizeof (ioctld));

    return ret;
}

int
find_pxe_sig(char *buf, int iter, struct sapiNBP *nbp)
{
    int len, i = 0;
    char *p = buf;
    char *sig = SIGNATURE;

    len = strlen(sig);
    while (INBP_BUF_SIZE - i >= len) {
	if (strncmp(p, sig, len) == 0) {
	    printf("\nGot the signature at 0x%x\n", iter * INBP_BUF_SIZE + i);
	    memcpy(nbp, p, sizeof (struct sapiNBP));
	    printf("targetname: %s\n\n", nbp->targetstring);
	    return 1;
	}
	p++;
	i++;
    }
    return 0;
}

void
get_inbp_info(struct sapiNBP *nbp)
{
    int fd, i = 0;
    char buf[1024];

    if ((fd = open("/dev/mem", O_RDONLY)) < 0) {
	perror("open mem");
	return;
    }
    while (i < SCANAREA) {
	if (read(fd, buf, INBP_BUF_SIZE) < 0) {
	    printf("Read error at 0x%x\n", i * INBP_BUF_SIZE);
	    perror("read");
	    return;
	}
	i++;
	/* printf("Reading %dk area..\n", i); */
	if (find_pxe_sig(buf, i, nbp))
	    break;
    }

    if (i > SCANAREA) {
	printf("\n%s not found, iscsi boot will fail !!!\n", SIGNATURE);
	sleep(10);
    }

}

void
establish_iscsi_network_boot_session(int bus_id, int target_id,
				     struct sapiNBP *nbp)
{
    int rv = 0;

    /* ensure the control device exists, and open it for all
     * of the child processes to use */
    for (;;) {
	control_fd = open_control_device();
	if (control_fd >= 0)
	    break;
	printf
	    ("could not open control device %s. Make sure iscsi_mod is loaded",
	     ISCSI_DEVICE);
	sleep(5);
    }

    do {
	rv = ioctl(control_fd, ISCSI_SET_INBP_INFO, (char *) nbp);
	if (rv == 0) {
	    printf("ISCSI_SET_INBP_INFO ioctl for %s successfully returned "
		   "from kernel with rv %d", nbp->targetstring, rv);
	} else {
	    printf("ISCSI_SET_INBP_INFO ioctl for %s failed, returned from "
		   "kernel with rv %d", nbp->targetstring, rv);
	}
    } while (rv != 0);

    do {
	rv = ioctl_establish_session(bus_id, target_id, nbp);
	if (rv == 0) {
	    printf("ISCSI_ESTABLISH_SESSION ioctl for %s successfully "
		   "returned from kernel with rv %d", nbp->targetstring, rv);
	} else {
	    printf("ISCSI_ESTABLISH_SESSION ioctl for %s failed, returned "
		   "from kernel with rv %d", nbp->targetstring, rv);
	}
    } while (rv != 0);
}
